/*
 * download.c: Web video plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <errno.h>
#include "download.h"
#include "common.h"

// --- cWebviThread --------------------------------------------------------

cWebviThread::cWebviThread(const char *templatePath) {
  webvi = webvi_initialize_context();

  if (webvi != 0 && templatePath)
    webvi_set_config(webvi, WEBVI_CONFIG_TEMPLATE_PATH, templatePath);
}

cWebviThread::~cWebviThread() {
  int numactive = activeRequestList.Size();
  for (int i=0; i<activeRequestList.Size(); i++)
    delete activeRequestList[i];
  activeRequestList.Clear();

  for (int i=0; i<finishedRequestList.Size(); i++) {
    delete finishedRequestList[i];
  }
  finishedRequestList.Clear();

  webvi_cleanup_context(webvi);

  if (numactive > 0) {
    esyslog("%d requests failed to complete", numactive);
  }

}

void cWebviThread::Action(void) {
  fd_set readfds, writefds, excfds;
  int maxfd;
  struct timeval timeout;
  long running_handles;
  WebviMsg *donemsg;
  int msg_remaining;
  cMenuRequest *req;

  if (webvi == 0) {
    error("Failed to get libwebvi context");
    return;
  }

  while (Running()) {

    requestMutex.Lock();
    for (int i=0; i<newRequestList.Size(); i++) {
      req = newRequestList[i];
      if (req->IsAborted()) {
        // The request has been aborted even before we got a chance to
        // start it.
	MoveToFinishedList(req);
      } else {
	debug("starting request %d", req->GetID());

        if (!req->Start(webvi)) {
	  error("Request failed to start");
	  req->RequestDone(-1, "Request failed to start");
	  MoveToFinishedList(req);
	} else {
	  activeRequestList.Append(req);
	}
      }
    }
    newRequestList.Clear();
    requestMutex.Unlock();

    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    FD_ZERO(&excfds);
    webvi_fdset(webvi, &readfds, &writefds, &excfds, &maxfd);

    timeout.tv_sec = 3;
    timeout.tv_usec = 0;

    int s = TEMP_FAILURE_RETRY(select(maxfd+1, &readfds, &writefds, NULL, 
                                      &timeout));
    if (s == -1) {
      // select error
      if (errno == EINTR)
        continue;

      LOG_ERROR_STR("select() error in downloader thread:");
      Cancel(-1);

    } else if (s == 0) {
      // timeout
      webvi_perform(webvi, 0, WEBVI_SELECT_TIMEOUT, &running_handles);

    } else {
      for (int fd=0; fd<=maxfd; fd++) {
        if (FD_ISSET(fd, &readfds))
          webvi_perform(webvi, fd, WEBVI_SELECT_READ, &running_handles);
        if (FD_ISSET(fd, &writefds))
          webvi_perform(webvi, fd, WEBVI_SELECT_WRITE, &running_handles);
        if (FD_ISSET(fd, &excfds))
          webvi_perform(webvi, fd, WEBVI_SELECT_EXCEPTION, &running_handles);
      }
    }

    do {
      donemsg = webvi_get_message(webvi, &msg_remaining);

      if (donemsg && donemsg->msg == WEBVIMSG_DONE) {
	requestMutex.Lock();
	req = activeRequestList.FindByHandle(donemsg->handle);
	if (req) {
	  req->RequestDone(donemsg->status_code, donemsg->data);
	  MoveToFinishedList(req);
	}
	requestMutex.Unlock();
      }
    } while (msg_remaining > 0);
  }
}

void cWebviThread::MoveToFinishedList(cMenuRequest *req) {
  // Move the request from the activeList to finishedList.
  requestMutex.Lock();
  for (int i=0; i<activeRequestList.Size(); i++) {
    if (activeRequestList[i] == req) {
      activeRequestList.Remove(i);
      break;
    }
  }
  finishedRequestList.Append(req);

  requestMutex.Unlock();
}

void cWebviThread::AddRequest(cMenuRequest *req) {
  requestMutex.Lock();
  newRequestList.Append(req);
  requestMutex.Unlock();
}

void cWebviThread::Abort(long downloadID) {
  cFileDownloadRequest *req = (cFileDownloadRequest *)downloadID;

  // Check if the request is still waiting to be started.
  bool found = false;
  requestMutex.Lock();
  for (int i=0; i<newRequestList.Size(); i++) {
    if (newRequestList[i] == req) {
      req->Abort();

      found = true;
      break;
    }
  }
  requestMutex.Unlock();

  if (found)
    return;

  // Check if the request is among the active ones.
  requestMutex.Lock();
  for (int i=0; i<activeRequestList.Size(); i++) {
    if (activeRequestList[i] == req) {
      req->Abort();
      break;
    }
  }
  requestMutex.Unlock();
}

cMenuRequest *cWebviThread::GetFinishedRequest() {
  cMenuRequest *res = NULL;
  requestMutex.Lock();
  if (finishedRequestList.Size() > 0) {
    res = finishedRequestList[finishedRequestList.Size()-1];
    finishedRequestList.Remove(finishedRequestList.Size()-1);
  }
  requestMutex.Unlock();

  return res;
}

int cWebviThread::GetUnfinishedCount() {
  if (!Running())
    return 0;
  else
    return activeRequestList.Size();
}

int cWebviThread::GetUnfinishedFilesSummary(cVector<cDownloadProgress *> *summaries) {
  cFileDownloadRequest *filereq;
  int nfilereq = 0;

  requestMutex.Lock();

  for (int i=0; i<activeRequestList.Size(); i++) {
    filereq = dynamic_cast<cFileDownloadRequest *> (activeRequestList[i]);
    if (filereq) {
      summaries->Append(filereq->GetProgress((long)filereq));
      nfilereq++;
    }
  }

  requestMutex.Unlock();

  return nfilereq;
}
